"""
RT Netlink protocol
"""

# 	Copyright (c) 2007 ALT Linux, Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 2 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from generic import *
from cxnet.common import *
import types

from os import listdir

##  RTnetlink multicast groups
RTNLGRP_NONE = 0x0
RTNLGRP_LINK = 0x1
RTNLGRP_NOTIFY = 0x2
RTNLGRP_NEIGH = 0x4
RTNLGRP_TC = 0x8
RTNLGRP_IPV4_IFADDR = 0x10
RTNLGRP_IPV4_MROUTE = 0x20
RTNLGRP_IPV4_ROUTE = 0x40
RTNLGRP_IPV4_RULE = 0x80
RTNLGRP_IPV6_IFADDR = 0x100
RTNLGRP_IPV6_MROUTE = 0x200
RTNLGRP_IPV6_ROUTE = 0x400
RTNLGRP_IPV6_IFINFO = 0x800
RTNLGRP_DECnet_IFADDR = 0x1000
RTNLGRP_NOP2 = 0x2000
RTNLGRP_DECnet_ROUTE = 0x4000
RTNLGRP_DECnet_RULE = 0x8000
RTNLGRP_NOP4 = 0x10000
RTNLGRP_IPV6_PREFIX = 0x20000
RTNLGRP_IPV6_RULE = 0x40000


## Types of messages
RTM_BASE	= 16
RTM_NEWLINK	= 16
RTM_DELLINK	= 17
RTM_GETLINK	= 18
RTM_SETLINK	= 19
RTM_NEWADDR	= 20
RTM_DELADDR	= 21
RTM_GETADDR	= 22
RTM_NEWROUTE	= 24
RTM_DELROUTE	= 25
RTM_GETROUTE	= 26
RTM_NEWNEIGH	= 28
RTM_DELNEIGH	= 29
RTM_GETNEIGH	= 30
RTM_NEWRULE	= 32
RTM_DELRULE	= 33
RTM_GETRULE	= 34
RTM_NEWQDISC	= 36
RTM_DELQDISC	= 37
RTM_GETQDISC	= 38
RTM_NEWTCLASS	= 40
RTM_DELTCLASS	= 41
RTM_GETTCLASS	= 42
RTM_NEWTFILTER	= 44
RTM_DELTFILTER	= 45
RTM_GETTFILTER	= 46
RTM_NEWACTION	= 48
RTM_DELACTION	= 49
RTM_GETACTION	= 50
RTM_NEWPREFIX	= 52
RTM_GETMULTICAST = 58
RTM_GETANYCAST	= 62
RTM_NEWNEIGHTBL	= 64
RTM_GETNEIGHTBL	= 66
RTM_SETNEIGHTBL = 67


class rtnl_hdr(Structure):
	_fields_ = [
		("length",	c_ushort),
		("type",	c_ushort),
	]

# 8<------------------------------------------------------------------------
class ifaddrmsg(Structure):
	_fields_ = [
		("family",	c_ubyte),	# Address family
		("prefixlen",	c_ubyte),	# Address' prefix length
		("flags",	c_ubyte),	# Address flags
		("scope",	c_ubyte),	# Adress scope
		("index",	c_int),		# Interface index
	]

class ifinfmsg(Structure):
	_fields_ = [
		("family",	c_ubyte),	# AF_UNSPEC (?)
		("type",	c_uint16),	# Interface type
		("index",	c_int),		# Interface index
		("flags",	c_int),		# Interface flags (netdevice(7))
		("change",	c_int),		# Change mask (reserved, always 0xFFFFFFFF)
	]

class ndmsg(Structure):
	_fields_ = [
		("family",	c_ubyte),	# 
		("index",	c_int),		# Interface index
		("state",	c_uint16),	# Neighbor entry state
		("flags",	c_uint8),	# Neighbor entry flags
		("type",	c_uint8),	#
	]

class rtmsg(Structure):
	_fields_ = [
		("family",	c_byte),	# Route address family
		("dst_len",	c_byte),	# Destination length (?)
		("src_len",	c_byte),	# Source path length (?)
		("tos",		c_byte),	# TOS filter
		("table",	c_byte),	# Routing table id
		("proto",	c_byte),	# Routing protocol
		("scope",	c_byte),
		("type",	c_byte),
		("flags",	c_int),
	]

# 8<------------------------------------------------------------------------

class rtnl_payload(Union):
	_fields_ = [
		("link",	ifinfmsg),
		("address",	ifaddrmsg),
		("route",	rtmsg),
		("neigh",	ndmsg),
		("raw",		(c_byte * NLMSG_MAX_LEN)),
	]

class rtnl_msg(Structure):
	_fields_ = [
		("hdr",		nlmsghdr),
		("data",	rtnl_payload),
	]


class rtnl_socket(nl_socket):
	"""
	Netlink RT socket implementation
	"""
	msg = rtnl_msg

	def __init__(self,groups = 0):
		nl_socket.__init__(self, family=NETLINK_ROUTE, groups=groups)


class rtnl_attr(object):
	"""
	RT Netlink message attribute
	"""

	def get_ip4ad_raw(self,address):
		return (c_uint8 * 4).from_address(address + sizeof(rtnl_hdr))

	def get_ip4ad(self,address):
		r = self.get_ip4ad_raw(address)
		return "%u.%u.%u.%u" % (r[0], r[1], r[2], r[3])

	def get_l2ad_raw(self,address):
		return (c_uint8 * 6).from_address(address + sizeof(rtnl_hdr))

	def get_l2ad(self,address):
		r = self.get_l2ad_raw(address)
		return "%x:%x:%x:%x:%x:%x" % (r[0], r[1], r[2], r[3], r[4], r[5])

	def get_uint(self,address):
		return c_uint.from_address(address + sizeof(rtnl_hdr)).value

	def get_asciiz(self,address):
		return string_at(address + sizeof(rtnl_hdr))

	def get_none(self,address):
		return None

## address attributes
#
# Important comment:
# IFA_ADDRESS is prefix address, rather than local interface address.
# It makes no difference for normally configured broadcast interfaces,
# but for point-to-point IFA_ADDRESS is DESTINATION address,
# local address is supplied in IFA_LOCAL attribute.
#
IFA_UNSPEC	= 0
IFA_ADDRESS	= 1
IFA_LOCAL	= 2
IFA_LABEL	= 3
IFA_BROADCAST	= 4
IFA_ANYCAST	= 5
IFA_CACHEINFO	= 6
IFA_MULTICAST	= 7

class ifa_attr (rtnl_attr):
	def __init__(self):
		self._map_ = {
			IFA_UNSPEC:	(self.get_none,		"none"),
			IFA_ADDRESS:	(self.get_ip4ad,	"address"),
			IFA_LOCAL:	(self.get_ip4ad,	"local"),
			IFA_LABEL:	(self.get_asciiz,	"dev"),
			IFA_BROADCAST:	(self.get_ip4ad,	"broadcast"),
			IFA_ANYCAST:	(self.get_ip4ad,	"anycast"),
			IFA_CACHEINFO:	(self.get_none,		"cacheinfo"),
			IFA_MULTICAST:	(self.get_ip4ad,	"multycast"),
		}

## neighbor attributes
NDA_UNSPEC	= 0
NDA_DST		= 1
NDA_LLADDR	= 2
NDA_CACHEINFO	= 3
NDA_PROBES	= 4

class nda_attr (rtnl_attr):
	def __init__(self):
		self._map_ = {
			NDA_UNSPEC:	(self.get_none,		"none"),
			NDA_DST:	(
						(self.get_ip4ad,	"dest"),
						(self.get_ip4ad_raw,	"raw_dest"),
					),
			NDA_LLADDR:	(
						(self.get_l2ad,		"lladdr"),
						(self.get_l2ad_raw,	"raw_lladdr"),
					),
			NDA_CACHEINFO:	(self.get_none,		"cacheinfo"),
			NDA_PROBES:	(self.get_none,		"probes"),
		}

## link attributes
IFLA_UNSPEC 	= 0
IFLA_ADDRESS	= 1
IFLA_BROADCAST	= 2
IFLA_IFNAME	= 3
IFLA_MTU	= 4
IFLA_LINK	= 5
IFLA_QDISC	= 6
IFLA_STATS	= 7
IFLA_COST	= 8
IFLA_PRIORITY	= 9
IFLA_MASTER	= 10
IFLA_WIRELESS	= 11 # Wireless Extension event - see iproute2:wireless.h 
IFLA_PROTINFO	= 12 # Protocol specific information for a link
IFLA_TXQLEN	= 13
IFLA_MAP	= 14
IFLA_WEIGHT	= 15
IFLA_OPERSTATE	= 16
IFLA_LINKMODE	= 17

class ifla_attr (rtnl_attr):
	def __init__(self):
		self._map_ = {
			IFLA_UNSPEC:	(self.get_none,		"none"),
			IFLA_ADDRESS:	(
						(self.get_l2ad,		"hwaddr"),
						(self.get_l2ad_raw,	"raw_hwaddr"),
					),
			IFLA_BROADCAST: (
						(self.get_l2ad,		"broadcast"),
						(self.get_l2ad_raw,	"raw_broadcast"),
					),
			IFLA_IFNAME:	(self.get_asciiz,	"dev"),
			IFLA_MTU:	(self.get_uint,		"mtu"),
			IFLA_LINK:	(self.get_uint,		"link"),
			IFLA_QDISC:	(self.get_asciiz,	"qdisc"),
			IFLA_STATS:	(self.get_none,		"stats"),
		}


## netdevice flags
iff = {}
iff["UP"]		= 0x1	# interface is up
iff["BROADCAST"]	= 0x2	# broadcast address valid
iff["DEBUG"]		= 0x4	# turn on debugging
iff["LOOPBACK"]		= 0x8	# is a loopback net
iff["POINTOPOINT"]	= 0x10	# interface is has p-p link
iff["NOTRAILERS"]	= 0x20	# avoid use of trailers
iff["RUNNING"]		= 0x40	# resources allocated
iff["NOARP"]		= 0x80	# no ARP protocol
iff["PROMISC"]		= 0x100	# receive all packets
iff["ALLMULTI"]		= 0x200	# receive all multicast packets
iff["MASTER"]		= 0x400	# master of a load balancer
iff["SLAVE"]		= 0x800	# slave of a load balancer
iff["MULTICAST"]	= 0x1000# supports multicast
iff["PORTSEL"]		= 0x2000# can set media type
iff["AUTOMEDIA"]	= 0x4000# auto media select active
iff["DYNAMIC"]		= 0x8000# dialup device with changing addresses



class rtnl_msg_parser(object):
	"""
	Generic RT Netlink attribute parser
	"""

	def get_attr(self,t,ptr,seen):
		hdr = rtnl_hdr.from_address(ptr)
		key = None
		result = None
		###
		# FIXME
		# x -- Alignment. Must NOT be hardcoded! But ctypes gives
		#      wrong alignment here.
		###
		x = 4
		k = hdr.length

		###
		# possible variants:
		#	* python-specific:	k - k%s + (k%x and x)
		#		cause in python ((y > 0) and x) == x
		#	* from lau:		((k + x - 1) / x) * x
		#	* from kernel sources:	(k + x - 1) & ~ (x - 1)
		align = (k + x - 1) & ~ (x - 1)

		if t._map_.has_key(hdr.type):
			if type(t._map_[hdr.type][0]) == types.TupleType:
				result = t._map_[hdr.type][seen][0](ptr)
				key    = t._map_[hdr.type][seen][1]
				seen  += 1
				if seen == len(t._map_[hdr.type]):
					seen = 0
				else:
					align  = 0
			else:
				result = t._map_[hdr.type][0](ptr)
				key    = t._map_[hdr.type][1]
				seen   = 0


		return (hdr.length,align,hdr.type,key,result,seen)

	def parse(self,msg):
		r = {}
		t = msg.hdr.type
		aa = [RTM_NEWADDR,RTM_NEWLINK,RTM_NEWROUTE,RTM_NEWNEIGH]
		ad = [RTM_DELADDR,RTM_DELLINK,RTM_DELROUTE,RTM_DELNEIGH]

		## message type
		if \
			t <= RTM_DELLINK:
			r["type"] = "link"
			r["index"] = msg.data.link.index
			r["flags"] = []
			for (i,k) in iff.items():
				if k & msg.data.link.flags:
					r["flags"].append(i)

			bias = ifinfmsg
			at = ifla_attr()
		elif \
			t <= RTM_DELADDR:
			r["type"] = "address"
			r["mask"] = msg.data.address.prefixlen
			bias = ifaddrmsg
			at = ifa_attr()
		elif \
			t <= RTM_GETNEIGH:
			r["type"] = "neigh"
			r["index"] = msg.data.neigh.index
			bias = ndmsg
			at = nda_attr()
		else:
			r["type"] = "fake"
			r["action"] = "fake"
			return r

		## message action
		if t in aa:
			r["action"] = "add"
		elif t in ad:
			r["action"] = "del"

		ptr = addressof(msg) + sizeof(nlmsghdr) + sizeof(bias)
		seen = 0

		while True:
			(k,l,t,key,data,seen) = self.get_attr(at,ptr,seen)
			if k == 0:
				break
			if key:
				r[key] = data


			ptr += l

		if r.has_key("dev"):
			###
			# find a PPP session for this device
			###
			if r["dev"][:3] == "ppp":
				# list all PPP session files in /var/run
				for i in listdir("/var/run"):
					if (i[:3] == "ppp") and (i[-3:] == "pid"):
						try:
							fd = open("/var/run/%s" % (i),"r")
							for m in fd.readlines():
								m = m.strip()
								if m == r["dev"]:
									r["session"] = i[4:-4]
							fd.close()
						except:
							pass
		return r


